@@ -1,35 +1,23 @@
 /*
- * Copyright 2019 Kai Uwe Broulik <kde@privat.broulik.de>
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) version 3, or any
- * later version accepted by the membership of KDE e.V. (or its
- * successor approved by the membership of KDE e.V.), which shall
- * act as a proxy defined in Section 6 of version 3 of the license.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library.  If not, see <http://www.gnu.org/licenses/>.
- */
+    SPDX-FileCopyrightText: 2019 Kai Uwe Broulik <kde@privat.broulik.de>
+
+    SPDX-License-Identifier: LGPL-2.1-only OR LGPL-3.0-only OR LicenseRef-KDE-Accepted-LGPL
+*/
 
 #pragma once
 
-#include <QSortFilterProxyModel>
 #include <QQmlParserStatus>
+#include <QSortFilterProxyModel>
+#include <QWindow>
 
-#include <QScopedPointer>
+#include <memory>
 
 #include "notificationmanager_export.h"
 
+#include <qqmlregistration.h>
+
 namespace NotificationManager
 {
-
 /**
  * @brief A model with notifications and jobs
  *
@@ -41,6 +29,7 @@ namespace NotificationManager
 class NOTIFICATIONMANAGER_EXPORT Notifications : public QSortFilterProxyModel, public QQmlParserStatus
 {
     Q_OBJECT
+    QML_ELEMENT
     Q_INTERFACES(QQmlParserStatus)
 
     /**
@@ -72,6 +61,25 @@ class NOTIFICATIONMANAGER_EXPORT Notifications : public QSortFilterProxyModel, p
      */
     Q_PROPERTY(bool showDismissed READ showDismissed WRITE setShowDismissed NOTIFY showDismissedChanged)
 
+    /**
+     * Whether to show notifications added during inhibition.
+     *
+     * If set to @c false, notifications are suppressed even after leaving "Do not disturb" mode.
+     *
+     * Default is @c true.
+     */
+    Q_PROPERTY(bool showAddedDuringInhibition READ showAddedDuringInhibition WRITE setShowAddedDuringInhibition NOTIFY showAddedDuringInhibitionChanged)
+
+    /**
+     * Whether to show blacklisted notifications added during inhibition.
+     *
+     * If set to @c true, notifications are shown even if blacklisted when they are added during inhibition.
+     *
+     * Default is @c false.
+     */
+    Q_PROPERTY(bool ignoreBlacklistDuringInhibition READ ignoreBlacklistDuringInhibition WRITE setIgnoreBlacklistDuringInhibition NOTIFY
+                   ignoreBlacklistDuringInhibitionChanged)
+
     /**
      * A list of desktop entries for which no notifications should be shown.
      *
@@ -206,6 +214,15 @@ class NOTIFICATIONMANAGER_EXPORT Notifications : public QSortFilterProxyModel, p
      */
     Q_PROPERTY(int unreadNotificationsCount READ unreadNotificationsCount NOTIFY unreadNotificationsCountChanged)
 
+    /**
+     * The number of resident notifications that have been dismissed
+     *
+     * This can be used to show that there is an active notification in the history
+     *
+     * @since 6.4
+     */
+    Q_PROPERTY(int dismissedResidentNotificationsCount READ dismissedResidentNotificationsCount NOTIFY dismissedResidentNotificationsCountChanged)
+
     /**
      * The number of active jobs
      */
@@ -218,6 +235,12 @@ class NOTIFICATIONMANAGER_EXPORT Notifications : public QSortFilterProxyModel, p
      */
     Q_PROPERTY(int jobsPercentage READ jobsPercentage NOTIFY jobsPercentageChanged)
 
+    /**
+     * The window that will render the notifications
+     *
+     * This is used to tell the xdg_activation_v1 protocol who is requesting the activation.
+     */
+    Q_PROPERTY(QWindow *window READ window WRITE setWindow NOTIFY windowChanged)
 public:
     explicit Notifications(QObject *parent = nullptr);
     ~Notifications() override;
@@ -238,14 +261,16 @@ public:
         UpdatedRole, ///< When the notification was last updated, invalid when it hasn't been updated.
 
         BodyRole, ///< The notification body text.
-        IconNameRole, ///< The notification main icon name, which is not the application icon. Only valid for icon names, if a URL supplied, it is loaded and exposed as ImageRole instead.
+        IconNameRole, ///< The notification main icon name, which is not the application icon. Only valid for icon names, if a URL supplied, it is loaded and
+                      ///< exposed as ImageRole instead.
 
         DesktopEntryRole, ///< The desktop entry (without .desktop suffix, e.g. org.kde.spectacle) of the application that sent the notification.
         NotifyRcNameRole, ///< The notifyrc name (e.g. spectaclerc) of the application that sent the notification.
 
         ApplicationNameRole, ///< The user-visible name of the application (e.g. Spectacle)
         ApplicationIconNameRole, ///< The icon name of the application
-        OriginNameRole, ///< The name of the device or account the notification originally came from, e.g. "My Phone" (in case of device sync) or "foo@example.com" (in case of an email notification)
+        OriginNameRole, ///< The name of the device or account the notification originally came from, e.g. "My Phone" (in case of device sync) or
+                        ///< "foo@example.com" (in case of an email notification)
 
         // Jobs
         JobStateRole, ///< The state of the job, either JobStateJopped, JobStateSuspended, or JobStateRunning.
@@ -263,15 +288,18 @@ public:
         UrlsRole, ///< A list of URLs associated with the notification, e.g. a path to a screenshot that was just taken or image received
 
         UrgencyRole, ///< The notification urgency, either LowUrgency, NormalUrgency, or CriticalUrgency. Jobs do not have an urgency.
-        TimeoutRole, ///< The timeout for the notification in milliseconds. 0 means the notification should not timeout, -1 means a sensible default should be applied.
+        TimeoutRole, ///< The timeout for the notification in milliseconds. 0 means the notification should not timeout, -1 means a sensible default should be
+                     ///< applied.
 
-        ConfigurableRole, ///< Whether the notification can be configured because a desktopEntry or notifyRcName is known, or the notification has a setting action. @sa configure
+        ConfigurableRole, ///< Whether the notification can be configured because a desktopEntry or notifyRcName is known, or the notification has a setting
+                          ///< action. @sa configure
         ConfigureActionLabelRole, ///< The user-visible label for the settings action
         ClosableRole, ///< Whether the item can be closed. Notifications are always closable, jobs are only when in JobStateStopped.
 
         ExpiredRole, ///< The notification timed out and closed. Actions on it cannot be invoked anymore.
         DismissedRole, ///< The notification got temporarily hidden by the user but could still be interacted with.
-        ReadRole, ///< Whether the notification got read by the user. If true, the notification isn't considered unread even if created after lastRead. @since 5.17
+        ReadRole, ///< Whether the notification got read by the user. If true, the notification isn't considered unread even if created after lastRead.
+                  ///< @since 5.17
 
         UserActionFeedbackRole, ///< Whether this notification is a response/confirmation to an explicit user action. @since 5.18
 
@@ -280,6 +308,15 @@ public:
         ReplyPlaceholderTextRole, ///< A custom placeholder text for the reply action, e.g. "Reply to Max...". @since 5.18
         ReplySubmitButtonTextRole, ///< A custom text for the reply submit button, e.g. "Submit Comment". @since 5.18
         ReplySubmitButtonIconNameRole, ///< A custom icon name for the reply submit button. @since 5.18
+        CategoryRole, ///< The (optional) category of the notification. Notifications can optionally have a type indicator. Although neither client or nor
+                      ///< server must support this, some may choose to. Those servers implementing categories may use them to intelligently display the
+                      ///< notification in a certain way, or group notifications of similar types.  @since 5.21
+        ResidentRole, ///< Whether the notification should keep its actions even when they were invoked. @since 5.22
+        TransientRole, ///< Whether the notification is transient and should not be kept in history. @since 5.22
+
+        WasAddedDuringInhibitionRole, ///< Whether the notification was added while inhibition was active. @since 6.3
+        HintsRole, ///< To provide extra data to a notification server that the server may be able to make use of. @since 6.4
+        DismissableRole, ///< Whether a notification can be dismissed (minimized). @since 6.4
     };
     Q_ENUM(Roles)
 
@@ -289,7 +326,7 @@ public:
     enum Type {
         NoType,
         NotificationType, ///< This item represents a notification.
-        JobType ///< This item represents an application job.
+        JobType, ///< This item represents an application job.
     };
     Q_ENUM(Type)
 
@@ -302,7 +339,7 @@ public:
         // these don't match the spec's value
         LowUrgency = 1 << 0, ///< The notification has low urgency, it is not important and may not be shown or added to a history.
         NormalUrgency = 1 << 1, ///< The notification has normal urgency. This is also the default if no urgecny is supplied.
-        CriticalUrgency = 1 << 2
+        CriticalUrgency = 1 << 2,
     };
     Q_ENUM(Urgency)
     Q_DECLARE_FLAGS(Urgencies, Urgency)
@@ -323,9 +360,9 @@ public:
      * The state an application job is in.
      */
     enum JobState {
-        JobStateStopped, ///< The job is stopped. It has either finished (error is 0) or failed (error is not 0)
+        JobStateStopped, ///< The job is stopped. It has either finished (error is 0) or failed (error is not 0).
         JobStateRunning, ///< The job is currently running.
-        JobStateSuspended ///< The job is currentl paused
+        JobStateSuspended, ///< The job is currently paused.
     };
     Q_ENUM(JobState)
 
@@ -335,7 +372,7 @@ public:
     enum SortMode {
         SortByDate = 0, ///< Sort notifications strictly by the date they were updated or created.
         // should this be flags? SortJobsFirst | SortByUrgency | ...?
-        SortByTypeAndUrgency ///< Sort notifications taking into account their type and urgency. The order is (descending): Critical, jobs, Normal, Low.
+        SortByTypeAndUrgency, ///< Sort notifications taking into account their type and urgency. The order is (descending): Critical, jobs, Normal, Low.
     };
     Q_ENUM(SortMode)
 
@@ -344,11 +381,19 @@ public:
      */
     enum GroupMode {
         GroupDisabled = 0,
-        //GroupApplicationsTree, // TODO make actual tree
-        GroupApplicationsFlat
+        // GroupApplicationsTree, // TODO make actual tree
+        GroupApplicationsFlat,
     };
     Q_ENUM(GroupMode)
 
+    enum InvokeBehavior {
+        None = 0,
+        Close = 1,
+    };
+    Q_ENUM(InvokeBehavior)
+    Q_DECLARE_FLAGS(InvokeBehaviors, InvokeBehavior)
+    Q_FLAG(InvokeBehaviors)
+
     int limit() const;
     void setLimit(int limit);
 
@@ -358,6 +403,12 @@ public:
     bool showDismissed() const;
     void setShowDismissed(bool show);
 
+    bool showAddedDuringInhibition() const;
+    void setShowAddedDuringInhibition(bool show);
+
+    bool ignoreBlacklistDuringInhibition() const;
+    void setIgnoreBlacklistDuringInhibition(bool show);
+
     QStringList blacklistedDesktopEntries() const;
     void setBlacklistedDesktopEntries(const QStringList &blacklist);
 
@@ -394,6 +445,9 @@ public:
     bool expandUnread() const;
     void setExpandUnread(bool expand);
 
+    QWindow *window() const;
+    void setWindow(QWindow *window);
+
     int count() const;
 
     int activeNotificationsCount() const;
@@ -404,6 +458,7 @@ public:
     void resetLastRead();
 
     int unreadNotificationsCount() const;
+    int dismissedResidentNotificationsCount() const;
 
     int activeJobsCount() const;
     int jobsPercentage() const;
@@ -445,7 +500,7 @@ public:
      * Invokes the action that should be triggered when clicking
      * the notification bubble itself.
      */
-    Q_INVOKABLE void invokeDefaultAction(const QModelIndex &idx);
+    Q_INVOKABLE void invokeDefaultAction(const QModelIndex &idx, InvokeBehavior behavior = None);
     /**
      * @brief Invoke a notification action
      *
@@ -453,7 +508,7 @@ public:
      * For invoking the default action, i.e. the one that is triggered
      * when clicking the notification bubble, use invokeDefaultAction
      */
-    Q_INVOKABLE void invokeAction(const QModelIndex &idx, const QString &actionId);
+    Q_INVOKABLE void invokeAction(const QModelIndex &idx, const QString &actionId, InvokeBehavior = None);
 
     /**
      * @brief Reply to a notification
@@ -461,7 +516,7 @@ public:
      * Replies to the given notification with the given text.
      * @since 5.18
      */
-    Q_INVOKABLE void reply(const QModelIndex &idx, const QString &text);
+    Q_INVOKABLE void reply(const QModelIndex &idx, const QString &text, InvokeBehavior behavior);
 
     /**
      * @brief Start automatic timeout of notifications
@@ -513,6 +568,20 @@ public:
 
     Q_INVOKABLE void collapseAllGroups();
 
+    Q_INVOKABLE void playSoundHint(const QModelIndex &idx) const;
+
+    /**
+     * Shows a notification to report the number of unread inhibited notifications.
+     */
+    Q_INVOKABLE void showInhibitionSummary(Urgency urgency, const QStringList &blacklistedDesktopEntries, const QStringList &blacklistedNotifyRcNames);
+
+    /**
+     * Hides the notification about the number of unread inhibited notifications.
+     *
+     * Call this when opening the notification history since the user will now be looking at the unread ones.
+     */
+    Q_INVOKABLE void hideInhibitionSummary();
+
     QVariant data(const QModelIndex &index, int role) const override;
     bool setData(const QModelIndex &index, const QVariant &value, int role) override;
     int rowCount(const QModelIndex &parent = QModelIndex()) const override;
@@ -521,10 +590,12 @@ public:
     bool filterAcceptsRow(int source_row, const QModelIndex &source_parent) const override;
     bool lessThan(const QModelIndex &source_left, const QModelIndex &source_right) const override;
 
-signals:
+Q_SIGNALS:
     void limitChanged();
     void showExpiredChanged();
     void showDismissedChanged();
+    void showAddedDuringInhibitionChanged();
+    void ignoreBlacklistDuringInhibitionChanged();
     void blacklistedDesktopEntriesChanged();
     void blacklistedNotifyRcNamesChanged();
     void whitelistedDesktopEntriesChanged();
@@ -542,8 +613,17 @@ signals:
     void expiredNotificationsCountChanged();
     void lastReadChanged();
     void unreadNotificationsCountChanged();
+    void dismissedResidentNotificationsCountChanged();
     void activeJobsCountChanged();
     void jobsPercentageChanged();
+    void windowChanged(QWindow *window);
+
+    /**
+     * Emitted when the user has requested to show the notifications popup.
+     *
+     * This is typically connected to a button in the "Missed Notifications in Do Not Disturb" notification.
+     */
+    void showNotificationsRequested();
 
 protected:
     void classBegin() override;
@@ -551,8 +631,7 @@ protected:
 
 private:
     class Private;
-    QScopedPointer<Private> d;
-
+    std::unique_ptr<Private> d;
 };
 
 } // namespace NotificationManager
